import numpy as np

# Global variables
time_step = 12.5
length_x = 119.70000
length_y = 4.67650      
length_z = 45.00000

"""
    ##############################################################
    ### From strata program (courtesy of Dr. Petter Johansson) ###
    ###       https://github.com/pjohansson/strata.git         ###
    ##############################################################
"""

FIELDS = ['X', 'Y', 'N', 'T', 'M', 'U', 'V']

def read_data(filename, decimals=5):

    """
    Read field data from a file name.

    Determines which of the simple formats in this module to use and
    returns data read using the proper function.

    Coordinates are rounded to input number of decimals.

    Args:
        filename (str): A file to read data from.

    Keyword Args:
        decimals (int): Number of decimals for coordinates.

    Returns:
        (dict, dict): 2-tuple of dict's with data and information. See
            strata.dataformats.read.read_data_file for more information.
    """
    
    with open(filename, 'rb') as fp:
        fields, num_values, info = read_header(fp)
        data = read_values(fp, num_values, fields)

    x0, y0 = info['origin']
    nx, ny = info['shape']
    dx, dy = info['spacing']

    x = x0 + dx * (np.arange(nx) + 0.5)
    y = y0 + dy * (np.arange(ny) + 0.5)
    xs, ys = np.meshgrid(x, y, indexing='ij')

    grid = np.zeros((nx, ny), dtype=[(l, np.float) for l in FIELDS])
    grid['X'] = xs
    grid['Y'] = ys

    for l in ['N', 'T', 'M', 'U', 'V']:
        grid[l][data['IX'], data['IY']] = data[l]

    grid = grid.ravel()

    return {l: grid[l] for l in FIELDS}, info
    

def calc_information(X, Y):

    """
    Return a dict of system information calculated from input cell positions.

    Calculates system origin ('origin'), bin spacing ('spacing'), number of cells
    ('num_bins') and shape ('shape').

    Args:
        X, Y (array_like): Arrays with cell positions.

    Returns:
        dict: Information about system in dictionary.
    """

    def calc_shape(X, Y):
        data = np.zeros((len(X), ), dtype=[('X', np.float), ('Y', np.float)])
        data['X'] = X
        data['Y'] = Y

        data.sort(order=['Y', 'X'])

        y0 = data['Y'][0]
        nx = 1

        try:
            while np.abs(data['Y'][nx] - y0) < 1e-4:
                nx += 1
        except IndexError:
           pass

        ny = len(X) // nx

        return nx, ny

    def calc_spacing(X, Y, nx, ny):
        def calc_1d(xs, n):
            x0 = np.min(xs)
            x1 = np.max(xs)

            try:
                return (x1 - x0) / (n - 1)
            except:
                return 0.0

        dx = calc_1d(X, nx)
        dy = calc_1d(Y, ny)

        return dx, dy

    def calc_origin(X, Y):
        return X.min(), Y.min()

    if len(X) != len(Y):
        raise ValueError("Lengths of X and Y arrays not equal")

    nx, ny = calc_shape(X, Y)
    dx, dy = calc_spacing(X, Y, nx, ny)
    x0, y0 = calc_origin(X, Y)

    info = {
        'shape': (nx, ny),
        'spacing': (dx, dy),
        'num_bins': nx * ny,
        'origin': (x0, y0),
    }

    return info

def read_binsimple(filename):

    """
    Return data and information read from a simple binary format.

    Args:
        filename (str): A file to read data from.

    Returns:
        dict: Data with field labels as keys.
    """

    def read_file(filename):
        # Fixed field order of format
        fields = ['X', 'Y', 'N', 'T', 'M', 'U', 'V']
        raw_data = np.fromfile(filename, dtype='float32')

        # Unpack into dictionary
        data = {}
        stride = len(fields)
        for i, field in enumerate(fields):
            data[field] = raw_data[i::stride]

        return data

    data = read_file(filename)

    return data

def read_plainsimple(filename):

    """
    Return field data from a simple plaintext format.

    Args:
        filename (str): A file to read data from.

    Returns:
        dict: Data with field labels as keys.
    """

    def read_file(filename):
        raw_data = np.genfromtxt(filename, names=True)

        # Unpack into dictionary
        data = {}
        for field in raw_data.dtype.names:
            data[field] = raw_data[field]

        return data

    data = read_file(filename)

    return data

def read_values(fp, num_values, fields):
    dtypes = {
        'IX': np.uint64,
        'IY': np.uint64,
        'N': np.float32,
        'T': np.float32,
        'M': np.float32,
        'U': np.float32,
        'V': np.float32,
    }

    return {
        l: np.fromfile(fp, dtype=dtypes[l], count=num_values)
        for l in fields
    }

def read_header(fp):
    """
    Read header information and forward the pointer to the data."""

    def read_shape(line):
        return tuple(int(v) for v in line.split()[1:3])

    def read_spacing(line):
        return tuple(float(v) for v in line.split()[1:3])

    def read_num_values(line):
        return int(line.split()[1].strip())

    def parse_field_labels(line):
        return line.split()[1:]

    def read_header_string(fp):
        buf_size = 1024
        header_str = ""

        while True:
            buf = fp.read(buf_size)

            pos = buf.find(b'\0')

            if pos != -1:
                header_str += buf[:pos].decode("ascii")
                offset = buf_size - pos - 1
                fp.seek(-offset, 1)
                break
            else:
                header_str += buf.decode("ascii")

        return header_str

    info = {}
    header_str = read_header_string(fp)

    for line in header_str.splitlines():
        line_type = line.split(maxsplit=1)[0].upper()

        if line_type == "SHAPE":
            info['shape'] = read_shape(line)
        elif line_type == "SPACING":
            info['spacing'] = read_spacing(line)
        elif line_type == "ORIGIN":
            info['origin'] = read_spacing(line)
        elif line_type == "FIELDS":
            fields = parse_field_labels(line)
        elif line_type == "NUMDATA":
            num_values = read_num_values(line)

    info['num_bins'] = info['shape'][0] * info['shape'][1]

    return fields, num_values, info

"""
    Read input file containing density map
    Returns the density map in kg/m^3 and the bin size in x and z directions
"""
def read_density (
    filename
    ) :

    # print('[densmap] Reading binary file in -flow format')
    data, info = read_data(filename)
    Nx = info['shape'][0]
    Nz = info['shape'][1]
    density_array = np.array( data['M'] )
    density_array = density_array.reshape((Nx,Nz))

    hx = length_x/Nx
    hz = length_z/Nz

    # From u/nm^3 to kg/m^3
    density_array *= 1.66053906660

    return density_array, hx, hz

"""
    Read velocity field from .dat data
    Returns the velocity components along x and z in nm/ps and the bin size in x and z directions
"""
def read_velocity (
    filename
    ) :

    data, info = read_data(filename)
    Nx = info['shape'][0]
    Nz = info['shape'][1]
    vel_x = np.array( data['U'] )
    vel_z = np.array( data['V'] )
    vel_x = vel_x.reshape((Nx,Nz))
    vel_z = vel_z.reshape((Nx,Nz))

    hx = length_x/Nx
    hz = length_z/Nz

    return vel_x, vel_z, hx, hz

"""
    Read temperature field from .dat data
    Returns the temperature in Kelvin and the bin size in x and z directions
"""
def read_temperature (
    filename
    ) :

    data, info = read_data(filename)
    Nx = info['shape'][0]
    Nz = info['shape'][1]
    temp = np.array( data['T'] )
    temp = temp.reshape((Nx,Nz))

    hx = length_x/Nx
    hz = length_z/Nz

    return temp, hx, hz
